from camomilla.models import Menu
from camomilla.serializers.base import BaseModelSerializer


class MenuSerializer(BaseModelSerializer):
    def get_default_field_names(self, *args):
        field_names = super().get_default_field_names(*args)
        self.action = getattr(
            self, "action", self.context and self.context.get("action", "list")
        )
        if self.action == "list":
            return [f for f in field_names if f != "nodes"]
        return field_names

    class Meta:
        model = Menu
        fields = "__all__"



import json
import os
from io import BytesIO

import magic
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.db import models
from django.db.models.fields.related import ForeignObjectRel
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from PIL import Image

from camomilla.fields import JSONField
from camomilla.settings import THUMBNAIL_FOLDER, THUMBNAIL_HEIGHT, THUMBNAIL_WIDTH
from camomilla.storages.optimize import OptimizedStorage


class AbstractMediaFolder(models.Model):
    title = models.CharField(max_length=200)
    slug = models.SlugField(editable=False, max_length=200, blank=True, null=True)
    creation_date = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)
    path = models.TextField(blank=True, null=True, editable=False)
    updir = models.ForeignKey(
        "self",
        on_delete=models.CASCADE,
        related_name="child_folders",
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True

    def update_childs(self):
        for folder in self.child_folders.all():
            folder.save()

    def save(self, *args, **kwargs):
        self.slug = slugify(self.title)
        if self.updir:
            if self.updir.id == self.id:
                raise ValidationError({"updir": "Unvalid parent"})
            self.path = "{0}/{1}".format(self.updir.path, self.slug)
        else:
            self.path = "/{0}".format(self.slug)

        super().save(*args, **kwargs)
        self.update_childs()

    def __str__(self):
        return "[%s] %s" % (self.__class__.__name__, self.title)


class MediaFolder(AbstractMediaFolder):
    pass


class Media(models.Model):
    # Seo Attributes
    alt_text = models.CharField(max_length=200, blank=True, null=True)
    title = models.CharField(max_length=200, blank=True, null=True)
    description = models.TextField(blank=True, null=True)

    file = models.FileField(storage=OptimizedStorage())
    thumbnail = models.ImageField(
        upload_to=THUMBNAIL_FOLDER,
        max_length=500,
        null=True,
        blank=True,
    )
    created = models.DateTimeField(auto_now=True)
    size = models.IntegerField(default=0, blank=True, null=True, editable=False)
    mime_type = models.CharField(max_length=128, blank=True, null=True, editable=False)
    image_props = JSONField(default=dict, blank=True, editable=False)
    folder = models.ForeignKey(
        MediaFolder,
        null=True,
        blank=True,
        related_name="media_folder",
        on_delete=models.CASCADE,
    )

    @property
    def path(self):
        return "%s/%s" % (self.folder.path, self.name)

    @property
    def is_image(self):
        return self.mime_type and self.mime_type.startswith("image")

    def image_preview(self):
        if self.file:
            return mark_safe('<img src="{0}" />'.format(self.file.url))

    def image_thumb_preview(self):
        if self.thumbnail:
            return mark_safe('<img src="{0}" />'.format(self.thumbnail.url))

    image_preview.short_description = _("Preview")
    image_thumb_preview.short_description = _("Thumbnail")

    class Meta:
        ordering = ["-pk"]

    def regenerate_thumbnail(self):
        self._remove_thumbnail()
        if self.file:
            self._make_thumbnail()

    def get_foreign_fields(self):
        return [
            field.get_accessor_name()
            for field in self._meta.get_fields()
            if issubclass(type(field), ForeignObjectRel)
        ]

    @property
    def json_repr(self):
        json_r = {
            "id": self.pk,
            "thumbnail": "" if not self.is_image else self.thumbnail.url,
            "label": self.__str__(),
        }
        return json.dumps(json_r)

    def _update_file_info(self, img_bytes=None):
        try:
            if not img_bytes:
                img_bytes = self.file.storage.open(self.file.name, "rb")
            self.mime_type = magic.from_buffer(img_bytes.read(2048), mime=True)
            with Image.open(img_bytes) as image:
                self.image_props = {
                    "width": image.width,
                    "height": image.height,
                    "format": image.format,
                    "mode": image.mode,
                }
        except Exception as ex:
            print(ex)
            return False

    def _make_thumbnail(self, img_bytes=None):
        try:
            if not img_bytes:
                img_bytes = self.file.storage.open(self.file.name, "rb")
            with Image.open(img_bytes) as orig_image:
                image = orig_image.copy()
                image.thumbnail((THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT), Image.ANTIALIAS)

                # Path to save to, name, and extension
                thumb_name, thumb_extension = os.path.splitext(self.file.name)
                thumb_extension = thumb_extension.lower()
                thumb_filename = thumb_name + "_thumb" + thumb_extension
                temp_thumb = BytesIO()
                image.save(temp_thumb, "PNG", optimize=True)
                temp_thumb.seek(0)
                # Load a ContentFile into the thumbnail field so it gets saved
                self.thumbnail.save(
                    thumb_filename, ContentFile(temp_thumb.read()), save=False
                )
                temp_thumb.close()
        except Exception:
            return False

        return True

    def _remove_file(self):
        if self.file:
            self.file.storage.delete(self.file.name)

    def _remove_thumbnail(self):
        if self.thumbnail:
            self.thumbnail.storage.delete(self.thumbnail.name)

    def _get_file_size(self):
        try:
            return self.file.storage.size(self.file.name)
        except Exception:
            return 0

    def __str__(self):
        if self.title:
            return self.title
        return self.file.name


@receiver(post_save, sender=Media, dispatch_uid="make thumbnails")
def update_media(sender, instance: Media, **kwargs):
    instance._remove_thumbnail()
    image_bytes = instance.file.storage.open(instance.file.name, "rb")
    instance._update_file_info(image_bytes)
    instance._make_thumbnail(image_bytes)
    Media.objects.filter(pk=instance.pk).update(
        size=instance._get_file_size(),
        thumbnail=instance.thumbnail,
        mime_type=instance.mime_type,
        image_props=instance.image_props,
    )


@receiver(pre_delete, sender=Media, dispatch_uid="make thumbnails")
def delete_media_files(sender, instance: Media, **kwargs):
    instance._remove_thumbnail()
    instance._remove_file()



from django import template
from camomilla.models import Menu

from camomilla.models.menu import MenuNode

register = template.Library()


@register.simple_tag(takes_context=True)
def get_menus(context, *args):
    if "menus" in context:
        return context["menus"]
    qs = Menu.objects.all()
    if len(args):
        qs = qs.filter(key__in=args)
    return Menu.defaultdict({m.key: m for m in qs})


@register.simple_tag(takes_context=True)
def render_menu(
    context: template.RequestContext,
    menu_key: str,
    template_path: str = "defaults/parts/menu.html",
):
    return context.get("menus", Menu.defaultdict())[menu_key].render(
        template_path=template_path,
        context=context,
        request=context.get("request", None),
    )


@register.filter(name="node_url")
def get_menu_node_url(node: MenuNode):
    return node.link.url



from typing import Any, Generic, Sequence
from typing_extensions import TypeVar, get_args
from django.core.exceptions import ImproperlyConfigured
from camomilla.utils.getters import pointed_getter


T = TypeVar("T")


class _LazyType:
    def __init__(self, path):
        self.path = path

    def evaluate(self, base_cls):
        module, type_name = self._evaluate_path(self.path, base_cls)
        return self._import(module, type_name)

    def _evaluate_path(self, relative_path, base_cls):
        base_module = base_cls.__module__

        modules = self._get_modules(relative_path, base_module)

        type_name = modules.pop()
        module = ".".join(modules)
        if not module:
            module = base_module
        return module, type_name

    def _get_modules(self, relative_path, base_module):
        canonical_path = relative_path.lstrip(".")
        canonical_modules = canonical_path.split(".")

        if not relative_path.startswith("."):
            return canonical_modules

        parents_amount = len(relative_path) - len(canonical_path)
        parent_modules = base_module.split(".")
        parents_amount = max(0, parents_amount - 1)
        if parents_amount > len(parent_modules):
            raise ValueError(f"Can't evaluate path '{relative_path}'")
        return parent_modules[: parents_amount * -1] + canonical_modules

    def _import(self, module_name, type_name):
        module = __import__(module_name, fromlist=[type_name])
        try:
            return getattr(module, type_name)
        except AttributeError:
            raise ValueError(f"Can't find type '{module_name}.{type_name}'.")


def get_type(source: Generic[T], raise_exception=True) -> T:
    try:
        return get_args(source)[0]
    except IndexError:
        if raise_exception:
            raise ImproperlyConfigured(
                "Must provide a Model class for ForeignKey fields."
            )
        else:
            return None


def get_type_eval(source: Generic[T], model: Any, raise_exception=True) -> T:
    type = get_type(source, raise_exception)
    if isinstance(type, str):
        return _LazyType(type).evaluate(model)


def set_key(data, key, val):
    if isinstance(data, Sequence):
        key = int(key)
        if key < len(data):
            data[key] = val
        else:
            data.append(val)
        return data
    elif isinstance(data, dict):
        data[key] = val
    else:
        setattr(data, key, val)
    return data


def get_key(data, key, default):
    if isinstance(data, Sequence):
        try:
            return data[int(key)]
        except IndexError:
            return default
    return pointed_getter(data, key, default)


def pointed_setter(data, path, value):
    path = path.split(".")
    key = path.pop(0)
    if not len(path):
        return set_key(data, key, value)
    default = [] if path[0].isdigit() else {}
    return set_key(
        data, key, pointed_setter(get_key(data, key, default), ".".join(path), value)
    )


def map_method_aliases(new_cls):
    method_aliases = {
        "validate_python": "model_validate",
        "validate_json": "model_validate_json",
        # "dump_python": "model_dump",
        # "dump_json": "model_dump_json",
        "json_schema": "model_json_schema",
    }
    for alias_name, target_name in method_aliases.items():
        setattr(new_cls, alias_name, getattr(new_cls, target_name))
    return new_cls



from django.shortcuts import redirect
from django.urls import include, path
from rest_framework import routers
from importlib.util import find_spec
from rest_framework.schemas import get_schema_view

from camomilla.openapi.schema import SchemaGenerator
from camomilla.views import (
    ArticleViewSet,
    CamomillaAuthLogin,
    CamomillaAuthLogout,
    CamomillaObtainAuthToken,
    ContentViewSet,
    LanguageViewSet,
    MediaFolderViewSet,
    MediaViewSet,
    PageViewSet,
    PermissionViewSet,
    TagViewSet,
    UserViewSet,
    MenuViewSet,
)
from camomilla.views.pages import fetch_page

router = routers.DefaultRouter()

router.register(r"tags", TagViewSet, "camomilla-tags")
router.register(r"articles", ArticleViewSet, "camomilla-articles")
router.register(r"contents", ContentViewSet, "camomilla-content")
router.register(r"media", MediaViewSet, "camomilla-media")
router.register(r"media-folders", MediaFolderViewSet, "camomilla-media_folders")
router.register(r"pages", PageViewSet, "camomilla-pages")
router.register(r"sitemap", PageViewSet, "camomilla-sitemap")
router.register(r"users", UserViewSet, "camomilla-users")
router.register(r"permissions", PermissionViewSet, "camomilla-permissions")
router.register(r"menus", MenuViewSet, "camomilla-menus")

urlpatterns = [
    path("", include(router.urls)),
    path("pages-router/", fetch_page),
    path("pages-router/<path:permalink>", fetch_page),
    path(
        "profiles/me/", lambda _: redirect("../../users/current/"), name="profiles-me"
    ),
    path("token-auth/", CamomillaObtainAuthToken.as_view(), name="api_token"),
    path("auth/login/", CamomillaAuthLogin.as_view(), name="login"),
    path("auth/logout/", CamomillaAuthLogout.as_view(), name="logout"),
    path("languages/", LanguageViewSet.as_view(), name="get_languages"),
    path(
        "openapi",
        get_schema_view(
            title="Camomilla",
            description="API for all things …",
            version="1.0.0",
            generator_class=SchemaGenerator,
        ),
        name="openapi-schema",
    ),
]

if find_spec("djsuperadmin.urls") is not None:
    urlpatterns += [
        path("djsuperadmin/", include("djsuperadmin.urls")),
    ]



import traceback
from io import BytesIO

from django.core.files.base import ContentFile
from django.core.files.storage import get_storage_class
from PIL import Image

from camomilla import settings


class OptimizedStorage(get_storage_class()):
    MEDIA_MAX_WIDTH = settings.MEDIA_OPTIMIZE_MAX_WIDTH
    MEDIA_MAX_HEIGHT = settings.MEDIA_OPTIMIZE_MAX_HEIGHT
    MEDIA_DPI = settings.MEDIA_OPTIMIZE_DPI
    MEDIA_OPTIMIZE_JPEG_QUALITY = settings.MEDIA_OPTIMIZE_JPEG_QUALITY

    def __init__(self, *args, **kwargs) -> None:
        self.MEDIA_MAX_WIDTH = int(kwargs.pop("max_width", self.MEDIA_MAX_WIDTH))
        self.MEDIA_MAX_HEIGHT = int(kwargs.pop("max_height", self.MEDIA_MAX_HEIGHT))
        self.MEDIA_DPI = int(kwargs.pop("dpi", self.MEDIA_DPI))
        self.MEDIA_OPTIMIZE_JPEG_QUALITY = int(
            kwargs.pop("jpeg_quality", self.MEDIA_OPTIMIZE_JPEG_QUALITY)
        )
        super().__init__(*args, **kwargs)

    def _save(self, name: str, content: ContentFile):
        if settings.ENABLE_MEDIA_OPTIMIZATION:
            content, _ = self._optimize(name, content)
        return super(OptimizedStorage, self)._save(name, content)

    def _optimize(self, name: str, content: ContentFile):
        try:
            image = Image.open(content)
            original_size = content.size
            width, height = image.size

            if width > self.MEDIA_MAX_WIDTH or height > self.MEDIA_MAX_HEIGHT:
                if width <= height:
                    selected_width = int((self.MEDIA_MAX_HEIGHT / height) * width)
                    selected_height = self.MEDIA_MAX_HEIGHT
                else:
                    selected_height = int((self.MEDIA_MAX_WIDTH / width) * height)
                    selected_width = self.MEDIA_MAX_WIDTH

                image = image.resize(
                    [selected_width, selected_height], resample=Image.LANCZOS
                )
            dpi = (self.MEDIA_DPI, self.MEDIA_DPI)
            image.info["dpi"] = dpi
            tmp = BytesIO()
            ext = name.split(".")[-1].lower().replace("jpg", "jpeg")
            image.save(
                tmp,
                ext,
                dpi=dpi,
                optimize=True,
                quality=self.MEDIA_OPTIMIZE_JPEG_QUALITY,
            )
            tmp_size = len(tmp.getvalue())

            if tmp_size > original_size:
                tmp.close()
                return content, False
            optimized_content = ContentFile(tmp.getvalue())
            tmp.close()
            content.close()
            return optimized_content, True
        except Exception as e:
            traceback.print_exc()
            print(f"Error optimizing image {name}: {str(e)}")
            return content, False



import json
from typing import Any, Union, TYPE_CHECKING

from django.db.models import JSONField
from django.db.models.query_utils import DeferredAttribute
from django_jsonform.models.fields import JSONFormField
from pydantic import (
    TypeAdapter,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
)

from .fields import *
from .models import *

if TYPE_CHECKING:
    from pydantic import BaseModel
    from pydantic.fields import Field
    from typing_extensions import Annotated


class StructuredJSONFormField(JSONFormField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = kwargs.get("encoder", None)
        self.decoder = kwargs.get("decoder", None)

    def prepare_value(self, value):
        if isinstance(value, list):
            return json.dumps([v.model_dump() for v in value])
        return value.model_dump_json()


class StructuredDescriptior(DeferredAttribute):
    field: "StructuredJSONField"

    def __set__(self, instance, value):
        # TODO: check if it's better to validate here or in __get__ function (performance reasons)
        # if not self.field.check_type(value):
        #     value = self.field.schema.validate_python(value)
        instance.__dict__[self.field.attname] = value

    def __get__(self, instance, cls=None):
        value = super().__get__(instance, cls)
        if not self.field.check_type(value):
            return self.field.schema.validate_python(value)
        return value


class StructuredJSONField(JSONField):
    # TODO: share cache in querysets of models having this same field
    # TODO: write queries for prefetch related for models inside the field

    descriptor_class = StructuredDescriptior

    @property
    def list_data_validator(self):
        def list_data_validator(
            value: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
        ) -> Any:
            from camomilla.structured.cache import CacheBuilder

            cache = CacheBuilder.from_model(self.orig_schema)
            if info.mode == "json" and isinstance(value, str):
                return self.schema.validate_python(
                    cache.inject_cache(json.loads(value))
                )
            return handler(cache.inject_cache(value))

        return list_data_validator

    def __init__(self, schema: type[BaseModel], *args: Any, **kwargs: Any) -> None:
        self.orig_schema = schema
        self.schema = schema
        default = kwargs.get("default", dict)
        self.file_handler = kwargs.pop("file_handler", "")
        self.many = kwargs.pop(
            "many", isinstance(default() if callable(default) else default, list)
        )
        if self.many:
            self.schema = TypeAdapter(
                Annotated[
                    list[self.schema],
                    Field(default_factory=list),
                    WrapValidator(self.list_data_validator),
                ]
            )
        return super().__init__(*args, **kwargs)

    def check_type(self, value: Any):
        if self.many:
            return isinstance(value, list) and all(
                isinstance(v, self.orig_schema) for v in value
            )
        return isinstance(value, self.orig_schema)

    def get_prep_value(
        self, value: Union[list[type[BaseModel]], type[BaseModel]]
    ) -> str:
        if isinstance(value, list) and self.many:
            return self.schema.dump_json(value, exclude_unset=True).decode()
        return value.model_dump_json(exclude_unset=True)

    def from_db_value(self, value: Any, expression: Any, connection: Any) -> Any:
        data = super().from_db_value(value, expression, connection)
        if isinstance(data, str):
            return json.loads(data)
        return data

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["schema"] = self.orig_schema
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": StructuredJSONFormField,
                "schema": self.schema.json_schema(),
                "model_name": self.model.__name__,
                "file_handler": self.file_handler,
                **kwargs,
            }
        )



def autodiscover_context_files():
    """
    Auto-discover INSTALLED_APPS template_context.py modules and fail silently when
    not present. This forces an import on them to register.
    Also import explicit modules.
    """
    import copy
    import sys
    from django.utils.module_loading import module_has_submodule
    from camomilla.templates_context.rendering import ctx_registry
    from importlib import import_module
    from django.apps import apps
    from camomilla.settings import TEMPLATE_CONTEXT_FILES, DEBUG

    mods = [
        (app_config.name, app_config.module) for app_config in apps.get_app_configs()
    ]

    for app, mod in mods:
        module = "%s.template_context" % app
        _model_registry_bkp = copy.copy(ctx_registry._model_registry)
        _template_registry_bkp = copy.copy(ctx_registry._template_registry)

        try:
            import_module(module)
        except Exception:
            ctx_registry._model_registry = _model_registry_bkp
            ctx_registry._template_registry = _template_registry_bkp
            if module_has_submodule(mod, "template_context"):
                raise

    for module in TEMPLATE_CONTEXT_FILES:
        import_module(module)

    if DEBUG:
        try:
            if sys.argv[1] in ("runserver", "runserver_plus"):
                if not ctx_registry.get_registered_info().items():
                    return
                print("Camomilla context files:")
                for k, v in ctx_registry.get_registered_info().items():
                    print(f"{k}:")
                    models = v.get("models")
                    templates = v.get("templates")
                    if models:
                        print(f"  models: {models}")
                    if templates:
                        print(f"  templates: {templates}")
                    print("\n")
        except IndexError:
            pass



from typing import Sequence, Tuple
from uuid import uuid4

from django.core.exceptions import ObjectDoesNotExist
from django.db import ProgrammingError, OperationalError, models, transaction
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.http import Http404
from django.urls import NoReverseMatch, reverse
from django.utils import timezone
from django.utils.functional import lazy
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from camomilla.models.mixins import MetaMixin, SeoMixin
from camomilla.utils import (
    activate_languages,
    get_field_translations,
    get_nofallbacks,
    lang_fallback_query,
    set_nofallbacks,
    url_lang_decompose,
    get_all_templates_files,
)
from camomilla.utils.getters import pointed_getter
from camomilla import settings
from camomilla.templates_context.rendering import ctx_registry


def GET_TEMPLATE_CHOICES():
    return [(t, t) for t in get_all_templates_files()]


class UrlNodeManager(models.Manager):
    @property
    def related_names(self):
        self._related_names = getattr(
            self,
            "_related_names",
            super().get_queryset().values_list("related_name", flat=True).distinct(),
        )
        return self._related_names

    def _annotate_fields(
        self,
        qs: models.QuerySet,
        field_names: Sequence[Tuple[str, models.Field, models.Value]],
    ):
        for field_name, output_field, default in field_names:
            whens = [
                models.When(
                    related_name=related_name,
                    then=models.F("__".join([related_name, field_name])),
                )
                for related_name in self.related_names
            ]
            qs = qs.annotate(
                **{
                    field_name: models.Case(
                        *whens, output_field=output_field, default=default
                    )
                }
            )
        return self._annotate_is_public(qs)

    def _annotate_is_public(self, qs: models.QuerySet):
        return qs.annotate(
            is_public=models.Case(
                models.When(status="PUB", then=True),
                models.When(
                    status="PLA", publication_date__lte=timezone.now(), then=True
                ),
                default=False,
                output_field=models.BooleanField(default=False),
            )
        )

    def get_queryset(self):
        try:
            return self._annotate_fields(
                super().get_queryset(),
                [
                    (
                        "indexable",
                        models.BooleanField(),
                        models.Value(None, models.BooleanField()),
                    ),
                    (
                        "status",
                        models.CharField(),
                        models.Value("DRF", models.CharField()),
                    ),
                    (
                        "publication_date",
                        models.DateTimeField(),
                        models.Value(timezone.now(), models.DateTimeField()),
                    ),
                    (
                        "date_updated_at",
                        models.DateTimeField(),
                        models.Value(timezone.now(), models.DateTimeField()),
                    )
                ],
            )
        except (ProgrammingError, OperationalError):
            return super().get_queryset()


class UrlNode(models.Model):
    permalink = models.CharField(max_length=400, unique=True, null=True)
    related_name = models.CharField(max_length=200)
    objects = UrlNodeManager()

    @property
    def page(self) -> "AbstractPage":
        return getattr(self, self.related_name)

    @staticmethod
    def reverse_url(permalink: str) -> str:
        try:
            if permalink == "/":
                return reverse("camomilla-homepage")
            return reverse("camomilla-permalink", args=(permalink.lstrip("/"),))
        except NoReverseMatch:
            return None

    @property
    def routerlink(self) -> str:
        return self.reverse_url(self.permalink) or self.permalink

    def get_absolute_url(self) -> str:
        return self.routerlink


PAGE_CHILD_RELATED_NAME = "%(app_label)s_%(class)s_child_pages"
URL_NODE_RELATED_NAME = "%(app_label)s_%(class)s"

PAGE_STATUS = (
    ("PUB", _("Published")),
    ("DRF", _("Draft")),
    ("TRS", _("Trash")),
    ("PLA", _("Planned")),
)


class PageBase(models.base.ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        attr_meta = attrs.pop("PageMeta", None)
        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        page_meta = attr_meta or getattr(new_class, "PageMeta", None)
        base_page_meta = getattr(new_class, "_page_meta", None)
        if page_meta:
            for name, value in getattr(base_page_meta, "__dict__", {}).items():
                if name not in page_meta.__dict__:
                    setattr(page_meta, name, value)
            setattr(new_class, "_page_meta", page_meta)
        return new_class


class AbstractPage(SeoMixin, MetaMixin, models.Model, metaclass=PageBase):
    date_created = models.DateTimeField(auto_now_add=True)
    date_updated_at = models.DateTimeField(auto_now=True)
    url_node = models.OneToOneField(
        UrlNode,
        on_delete=models.CASCADE,
        related_name=URL_NODE_RELATED_NAME,
        null=True,
        editable=False,
    )
    breadcrumbs_title = models.CharField(max_length=128, null=True, blank=True)
    slug = models.SlugField(max_length=150, allow_unicode=True, null=True, blank=True)
    status = models.CharField(
        max_length=3,
        choices=PAGE_STATUS,
        default="DRF",
    )
    template = models.CharField(max_length=500, null=True, blank=True, choices=[])
    template_data = models.JSONField(default=dict, null=False, blank=True)
    identifier = models.CharField(max_length=200, null=True, unique=True, default=uuid4)
    publication_date = models.DateTimeField(null=True, blank=True)
    indexable = models.BooleanField(default=True)
    ordering = models.PositiveIntegerField(default=0, blank=False, null=False)
    parent_page = models.ForeignKey(
        "self",
        related_name=PAGE_CHILD_RELATED_NAME,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )

    def __init__(self, *args, **kwargs):
        super(AbstractPage, self).__init__(*args, **kwargs)
        self._meta.get_field("template").choices = lazy(GET_TEMPLATE_CHOICES, list)()

    def __str__(self) -> str:
        return "(%s) %s" % (self.__class__.__name__, self.title or self.permalink)

    def get_context(self, request=None):
        context = {
            "page": self,
            "page_model": {"class": self.__class__.__name__, "module": self.__module__},
            "request": request,
        }
        inject_func = pointed_getter(self, "_page_meta.inject_context_func")
        if inject_func and callable(inject_func):
            new_ctx = inject_func(request=request, super_ctx=context)
            if isinstance(new_ctx, dict):
                context.update(new_ctx)
        return ctx_registry.get_context_for_page(self, request, super_ctx=context)

    @property
    def model_name(self) -> str:
        return self._meta.app_label + "." + self._meta.model_name

    @property
    def model_info(self) -> dict:
        return {"app_label": self._meta.app_label, "class": self._meta.model_name}

    @property
    def permalink(self) -> str:
        return self.url_node and self.url_node.permalink

    @property
    def routerlink(self) -> str:
        return self.url_node and self.url_node.routerlink

    @property
    def breadcrumbs(self) -> Sequence[dict]:
        breadcrumb = {
            "permalink": self.permalink,
            "title": self.breadcrumbs_title or self.title or self.slug,
        }
        if self.parent:
            return self.parent.breadcrumbs + [breadcrumb]
        return [breadcrumb]

    @property
    def is_public(self) -> bool:
        status = get_nofallbacks(self, "status")
        publication_date = get_nofallbacks(self, "publication_date")
        if status == "PUB":
            return True
        if status == "PLA":
            return (
                bool(publication_date) and timezone.now() > publication_date
            )
        return False

    def get_template_path(self, request=None) -> str:
        return self.template or pointed_getter(self, "_page_meta.default_template")

    @property
    def childs(self) -> models.Manager:
        if hasattr(self._page_meta, "child_page_field"):
            return getattr(self, self._page_meta.child_page_field)
        return getattr(self, PAGE_CHILD_RELATED_NAME % self.model_info)

    @property
    def parent(self) -> models.Model:
        return getattr(self, self._page_meta.parent_page_field)

    def _get_or_create_url_node(self) -> UrlNode:
        if not self.url_node:
            self.url_node = UrlNode.objects.create(
                related_name=URL_NODE_RELATED_NAME % self.model_info
            )
        return self.url_node

    def _update_url_node(self, force: bool = False) -> UrlNode:
        self.url_node = self._get_or_create_url_node()
        for __ in activate_languages():
            old_permalink = self.permalink
            new_permalink = self.generate_permalink()
            force = force or old_permalink != new_permalink
            set_nofallbacks(self.url_node, "permalink", new_permalink)
        if force:
            self.url_node.save()
            self.update_childs()
        return self.url_node

    def generate_permalink(self, safe: bool = True) -> str:
        slug = get_nofallbacks(self, "slug")
        if slug is None and not self.permalink:
            translations = get_field_translations(self, "slug").values()
            fallback_slug = next((t for t in translations if t is not None), None)
            slug = (
                slugify(self.title or uuid4(), allow_unicode=True)
                if fallback_slug is None
                else fallback_slug
            )
            set_nofallbacks(self, "slug", slug)
        permalink = "/%s" % slugify(slug or "", allow_unicode=True)
        if self.parent:
            permalink = f"{self.parent.permalink}{permalink}"
        qs = UrlNode.objects.exclude(pk=getattr(self.url_node or object, "pk", None))
        if safe and qs.filter(permalink=permalink).exists():
            permalink = "/".join(
                permalink.split("/")[:-1] + [slugify(uuid4(), allow_unicode=True)]
            )
        return permalink

    def update_childs(self) -> None:
        # without pk, no childs there
        if self.pk is not None:
            for child in self.childs.all():
                child.save()

    def save(self, *args, **kwargs) -> None:
        with transaction.atomic():
            self._update_url_node()
            return super().save(*args, **kwargs)

    @classmethod
    def get(cls, request, *args, **kwargs) -> "AbstractPage":
        bypass_type_check = kwargs.pop("bypass_type_check", False)
        bypass_public_check = kwargs.pop("bypass_public_check", False)

        if len(kwargs.keys()) > 0:
            page = cls.objects.get(**kwargs)
        else:
            node = UrlNode.objects.filter(
                permalink=url_lang_decompose(request.path)["permalink"]
            ).first()
            page = node and node.page
        type_error = not bypass_type_check and not isinstance(page, cls)
        public_error = not bypass_public_check and not getattr(
            page or object, "is_public", False
        )
        if not page or type_error or public_error:
            bases = (UrlNode.DoesNotExist,)
            if hasattr(cls, "DoesNotExist"):
                bases += (cls.DoesNotExist,)
            message = "%s matching query does not exist." % cls._meta.object_name
            if public_error:
                message = (
                    "Match found: %s.\nThe page appears not to be public.\nUse ?preview=true in the url to see it."
                    % page
                )
            raise type("PageDoesNotExist", bases, {})(message)
        return page

    @classmethod
    def get_or_create(cls, request, *args, **kwargs) -> Tuple["AbstractPage", bool]:
        try:
            return cls.get(request, *args, **kwargs), False
        except ObjectDoesNotExist:
            if len(kwargs.keys()) > 0:
                return cls.objects.get_or_create(**kwargs)
        return (None, False)

    @classmethod
    def get_or_create_homepage(cls) -> Tuple["AbstractPage", bool]:
        try:
            if settings.ENABLE_TRANSLATIONS:
                node = UrlNode.objects.get(lang_fallback_query(permalink="/"))
            else:
                node = UrlNode.objects.get(permalink="/")
            return node.page, False
        except UrlNode.DoesNotExist:
            return cls.get_or_create(None, slug="")

    @classmethod
    def get_or_404(cls, request, *args, **kwargs) -> "AbstractPage":
        try:
            return cls.get(request, *args, **kwargs)
        except ObjectDoesNotExist as ex:
            raise Http404(ex)

    def alternate_urls(self, *args, **kwargs) -> dict:
        permalinks = get_field_translations(self.url_node or object, "permalink", None)
        for lang in activate_languages():
            if lang in permalinks:
                permalinks[lang] = UrlNode.reverse_url(permalinks[lang])
        return permalinks

    class Meta:
        abstract = True
        ordering = ("ordering",)
        verbose_name = _("Page")
        verbose_name_plural = _("Pages")

    class PageMeta:
        parent_page_field = "parent_page"
        default_template = settings.PAGE_DEFAULT_TEMPLATE
        inject_context_func = settings.PAGE_INJECT_CONTEXT_FUNC


class Page(AbstractPage):
    pass


@receiver(post_delete)
def auto_delete_url_node(sender, instance, **kwargs):
    if issubclass(sender, AbstractPage):
        instance.url_node and instance.url_node.delete()



import json

import django
from django.conf import settings

if django.VERSION >= (4, 0):
    from django.db.models import JSONField as DjangoJSONField
else:
    from django.contrib.postgres.fields import JSONField as DjangoJSONField

from django.contrib.postgres.fields import ArrayField as DjangoArrayField
from django.db import models


if "sqlite" in settings.DATABASES["default"]["ENGINE"]:  # noqa: C901

    class JSONField(models.Field):
        def db_type(self, connection):
            return "text"

        def from_db_value(self, value, expression, connection):
            if value is not None:
                return self.to_python(value)
            return value

        def to_python(self, value):
            if value is not None:
                try:
                    return json.loads(value)
                except (TypeError, ValueError):
                    return value
            return value

        def get_prep_value(self, value):
            if value is not None:
                return str(json.dumps(value))
            return value

        def value_to_string(self, obj):
            return self.value_from_object(obj)

    class ArrayField(JSONField):
        def __init__(self, base_field, size=None, **kwargs):
            """Care for DjanroArrayField's kwargs."""
            self.base_field = base_field
            self.size = size
            return super().__init__(**kwargs)

        def deconstruct(self):
            """Need to create migrations properly."""
            name, path, args, kwargs = super().deconstruct()
            kwargs.update(
                {
                    "base_field": self.base_field.clone(),
                    "size": self.size,
                }
            )
            return name, path, args, kwargs

else:

    class JSONField(DjangoJSONField):
        pass

    class ArrayField(DjangoArrayField):
        pass


__all__ = [JSONField, ArrayField]



import re
from typing import Any, Sequence, Iterator

from django.db.models import Model, Q
from django.utils.translation.trans_real import activate, get_language
from modeltranslation.settings import AVAILABLE_LANGUAGES, DEFAULT_LANGUAGE
from modeltranslation.utils import build_localized_fieldname
from camomilla.settings import BASE_URL


def activate_languages(languages: Sequence[str] = AVAILABLE_LANGUAGES) -> Iterator[str]:
    old = get_language()
    for language in languages:
        activate(language)
        yield language
    activate(old)


def set_nofallbacks(instance: Model, attr: str, value: Any, **kwargs) -> None:
    language = kwargs.pop("language", get_language())
    local_fieldname = build_localized_fieldname(attr, language)
    if hasattr(instance, local_fieldname):
        attr = local_fieldname
    return setattr(instance, attr, value)


def get_nofallbacks(instance: Model, attr: str, *args, **kwargs) -> Any:
    language = kwargs.pop("language", get_language())
    local_fieldname = build_localized_fieldname(attr, language)
    if hasattr(instance, local_fieldname):
        attr = local_fieldname
    return getattr(instance, attr, *args, **kwargs)


def url_lang_decompose(url):
    if BASE_URL and url.startswith(BASE_URL):
        url = url[len(BASE_URL):]
    data = {"url": url, "permalink": url, "language": DEFAULT_LANGUAGE}
    result = re.match(
        f"^\/?({'|'.join(AVAILABLE_LANGUAGES)})?\/(.*)", url
    )  # noqa: W605
    groups = result and result.groups()
    if groups and len(groups) == 2:
        data["language"] = groups[0]
        data["permalink"] = "/%s" % groups[1]
    return data


def get_field_translations(instance: Model, field_name: str, *args, **kwargs):
    return {
        lang: get_nofallbacks(instance, field_name, language=lang, *args, **kwargs)
        for lang in AVAILABLE_LANGUAGES
    }


def lang_fallback_query(**kwargs):
    current_lang = get_language()
    query = Q()
    for lang in AVAILABLE_LANGUAGES:
        query |= Q(**{f"{key}_{lang}": value for key, value in kwargs.items()})
    if current_lang:
        query = query & Q(
            **{f"{key}_{current_lang}__isnull": True for key in kwargs.keys()}
        )
        query |= Q(**{f"{key}_{current_lang}": value for key, value in kwargs.items()})
    return query


def is_translatable(model: Model) -> bool:
    from modeltranslation.translator import translator

    return model in translator.get_registered_models()



from rest_framework import authentication


class SessionAuthentication(authentication.SessionAuthentication):
    """
    This class is needed, because REST Framework's default SessionAuthentication does never return 401's,
    because they cannot fill the WWW-Authenticate header with a valid value in the 401 response. As a
    result, we cannot distinguish calls that are not unauthorized (401 unauthorized) and calls for which
    the user does not have permission (403 forbidden). See https://github.com/encode/django-rest-framework/issues/5968

    We do set authenticate_header function in SessionAuthentication, so that a value for the WWW-Authenticate
    header can be retrieved and the response code is automatically set to 401 in case of unauthenticated requests.
    """

    def authenticate_header(self, request):
        return "Session"



from django.db import models

from camomilla.structured import StructuredJSONField

from .json import ArrayField, JSONField

ORDERING_ACCEPTED_FIELDS = (
    models.BigIntegerField,
    models.IntegerField,
    models.PositiveIntegerField,
    models.PositiveSmallIntegerField,
    models.SmallIntegerField,
)

__all__ = ["StructuredJSONField", "JSONField", "ArrayField", "ORDERING_ACCEPTED_FIELDS"]



from rest_framework import permissions


class CamomillaBasePermissions(permissions.BasePermission):
    def _check_custom_permissions(self, request, model):
        if request.method in permissions.SAFE_METHODS:
            return True
        else:
            permission = ""
            model_name = model.__name__.lower()
            if request.method == "GET":
                permission = "view"
            if request.method == "DELETE":
                permission = "delete"
            if request.method == "POST":
                permission = "add"
            if request.method in ["PUT", "PATCH"]:
                permission = "change"
            perm = "{0}.{1}_{2}".format(model._meta.app_label, permission, model_name)
            return request.user.has_perm(perm)

    def has_permission(self, request, view):
        if request.user.is_authenticated and request.user.is_superuser:
            return True
        if not request.user.is_authenticated:
            return False
        return self._check_custom_permissions(request, view.model)

    def has_object_permission(self, request, view, obj):
        if request.user.is_authenticated and request.user.is_superuser:
            return True
        if not request.user.is_authenticated:
            return False
        return self._check_custom_permissions(request, view.model)


class CamomillaSuperUser(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.user.is_authenticated and request.user.is_superuser

    def has_object_permission(self, request, view, obj):
        return request.user.is_authenticated and request.user.is_superuser


class ReadOnly(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.method in permissions.SAFE_METHODS



from typing import Any, Callable, Dict, Generic, TypeVar, Union

from django.db import models as django_models
from pydantic import SerializationInfo
from pydantic_core import core_schema as cs

from .utils import get_type


T = TypeVar("T", bound=django_models.Model)


class ForeignKey(Generic[T]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Any, handler: Callable[[Any], cs.CoreSchema]
    ) -> cs.CoreSchema:
        from camomilla.structured.cache import ValueWithCache

        model_class = get_type(source)

        def validate_from_pk(pk: Union[int, str]) -> model_class:
            return model_class._default_manager.get(pk=pk)

        int_str_union = cs.union_schema([cs.str_schema(), cs.int_schema()])
        from_pk_schema = cs.chain_schema(
            [
                int_str_union,
                cs.no_info_plain_validator_function(validate_from_pk),
            ]
        )
        pk_attname = model_class._meta.pk.attname

        def validate_from_dict(data: Dict[str, Union[str, int]]) -> model_class:
            return validate_from_pk(data[pk_attname])

        from_dict_schema = cs.chain_schema(
            [
                cs.typed_dict_schema({pk_attname: cs.typed_dict_field(int_str_union)}),
                cs.no_info_plain_validator_function(validate_from_dict),
            ]
        )

        from_cache_schema = cs.chain_schema(
            [
                cs.is_instance_schema(ValueWithCache),
                cs.no_info_plain_validator_function(lambda v: v.retrieve()),
            ]
        )

        def serialize_data(instance, info):
            from camomilla.serializers.utils import build_standard_model_serializer

            if info.mode == "python":
                serializer = build_standard_model_serializer(model_class, depth=1)
                return serializer(instance=instance).data
            return instance.pk

        return cs.json_or_python_schema(
            json_schema=cs.union_schema(
                [from_cache_schema, from_pk_schema, from_dict_schema]
            ),
            python_schema=cs.union_schema(
                [
                    cs.is_instance_schema(model_class),
                    from_cache_schema,
                    from_pk_schema,
                    from_dict_schema,
                ]
            ),
            serialization=cs.plain_serializer_function_ser_schema(
                serialize_data, info_arg=True
            ),
        )


class QuerySet(Generic[T]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Any, handler: Callable[[Any], cs.CoreSchema]
    ) -> cs.CoreSchema:
        from camomilla.structured.cache import ValueWithCache

        model_class = get_type(source)

        def validate_from_pk_list(
            values: list[Union[int, str]]
        ) -> django_models.QuerySet:
            preserved = django_models.Case(
                *[django_models.When(pk=pk, then=pos) for pos, pk in enumerate(values)]
            )
            return model_class._default_manager.filter(pk__in=values).order_by(
                preserved
            )

        int_str_union = cs.union_schema([cs.str_schema(), cs.int_schema()])
        from_pk_list_schema = cs.chain_schema(
            [
                cs.list_schema(int_str_union),
                cs.no_info_plain_validator_function(validate_from_pk_list),
            ]
        )
        pk_attname = model_class._meta.pk.attname

        def validate_from_dict(
            values: list[Dict[str, Union[str, int]]]
        ) -> django_models.QuerySet:
            return validate_from_pk_list([data[pk_attname] for data in values])

        from_dict_list_schema = cs.chain_schema(
            [
                cs.list_schema(
                    cs.typed_dict_schema(
                        {pk_attname: cs.typed_dict_field(int_str_union)}
                    )
                ),
                cs.no_info_plain_validator_function(validate_from_dict),
            ]
        )
        from_cache_schema = cs.chain_schema(
            [
                cs.is_instance_schema(ValueWithCache),
                cs.no_info_plain_validator_function(lambda v: v.retrieve()),
            ]
        )

        def serialize_data(qs: django_models.QuerySet, info: SerializationInfo):
            from camomilla.serializers.utils import build_standard_model_serializer

            if info.mode == "python":
                serializer = build_standard_model_serializer(model_class, depth=1)
                return serializer(instance=qs, many=True).data
            return [x.pk for x in qs]

        return cs.json_or_python_schema(
            json_schema=cs.union_schema(
                [from_cache_schema, from_pk_list_schema, from_dict_list_schema]
            ),
            python_schema=cs.union_schema(
                [
                    cs.is_instance_schema(django_models.QuerySet),
                    from_cache_schema,
                    from_pk_list_schema,
                    from_dict_list_schema,
                ]
            ),
            serialization=cs.plain_serializer_function_ser_schema(
                serialize_data, info_arg=True
            ),
        )



from pathlib import Path
from typing import Sequence

from django import template as django_template
from os.path import relpath


def get_all_templates_files() -> Sequence[str]:
    dirs = []
    for engine in django_template.loader.engines.all():
        # Exclude pip installed site package template dirs
        dirs.extend(
            x
            for x in engine.template_dirs
            if "site-packages" not in str(x) or "camomilla" in str(x)
        )
    files = []
    for dir in dirs:
        files.extend(relpath(x, dir) for x in Path(dir).glob("**/*.html") if x)
    return files



from ..mixins import OptimViewMixin, PaginateStackMixin, OrderingMixin
from rest_framework import viewsets


class BaseModelViewset(
    OptimViewMixin, OrderingMixin, PaginateStackMixin, viewsets.ModelViewSet
):
    pass



import os
import mock

from django.http import Http404
from django.test import TestCase
from django.test import RequestFactory
from django.utils.translation import activate, get_language

from camomilla.models import Page, Article
from camomilla.templatetags.camomilla_filters import filter_content, alternate_urls

from requests import RequestException


class CamomillaFiltersTestCase(TestCase):
    def setUp(self):
        pass

    def test_filter_content(self):
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        page = Page.get(request, identifier="home")
        content = filter_content(page, "content1")
        self.assertEqual(content.identifier, "content1")
        self.assertEqual(content.content, "")
        content.content = "Hello World!"
        content.save()
        page = Page.get(request, identifier="home")
        content = filter_content(page, "content1")
        self.assertEqual(content.identifier, "content1")
        self.assertEqual(content.content, "Hello World!")

    def test_filter_alternate_urls(self):
        request = RequestFactory().get("/path", HTTP_HOST="localhost:8000")
        request.META["HTTP_HOST"] = "localhost"
        page = Page.get(request, identifier="home")
        alt_urls = dict(alternate_urls(page, request))
        self.assertEqual(alt_urls, {})

        request = RequestFactory().get("/about", HTTP_HOST="localhost:8000")
        request.META["HTTP_HOST"] = "localhost"
        page = Page.get(request, identifier="about")
        alt_urls = dict(alternate_urls(page, request))
        self.assertEqual(alt_urls["it"], "http://localhost/about")
        self.assertEqual(alt_urls["en"], "http://localhost/en/about")
        self.assertEqual(alt_urls["de"], "http://localhost/de/about")



import django
from django.conf import settings as django_settings
from django.db.models.aggregates import Max
from django.db.models.functions import Coalesce
from django.utils import translation
from modeltranslation.settings import AVAILABLE_LANGUAGES
from modeltranslation.utils import build_localized_fieldname
from rest_framework import serializers
from rest_framework.utils import model_meta

from camomilla.fields import ORDERING_ACCEPTED_FIELDS
from camomilla.serializers.fields.related import RelatedField
from camomilla.serializers.utils import build_standard_model_serializer
from camomilla.serializers.validators import UniquePermalinkValidator
from camomilla.utils import dict_merge
from camomilla import settings

if django.VERSION >= (4, 0):
    from django.db.models import JSONField as DjangoJSONField
else:
    from django.contrib.postgres.fields import JSONField as DjangoJSONField

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from camomilla.models.page import AbstractPage


# TODO: decide what to do with LangInfoMixin mixin!
class LangInfoMixin(metaclass=serializers.SerializerMetaclass):
    """
    This mixin adds a "lang_info" field to the serializer.
    This field contains information about the languages available in the site.
    """
    lang_info = serializers.SerializerMethodField("get_lang_info", read_only=True)

    def get_lang_info(self, obj, *args, **kwargs):
        languages = []
        for key, language in django_settings.LANGUAGES:
            languages.append({"id": key, "name": language})
        return {
            "default": django_settings.LANGUAGE_CODE,
            "active": translation.get_language(),
            "site_languages": languages,
        }

    def get_default_field_names(self, *args):
        field_names = super().get_default_field_names(*args)
        self.action = getattr(
            self, "action", self.context and self.context.get("action", "list")
        )
        if self.action != "retrieve":
            return [f for f in field_names if f != "lang_info"]
        return field_names


class SetupEagerLoadingMixin:
    """
    This mixin allows to use the setup_eager_loading method to optimize the queries.
    """
    @staticmethod
    def setup_eager_loading(queryset):
        return queryset


class OrderingMixin:
    """
    This mixin allows to set the default value of an ordering field to the max value + 1.
    """

    def get_max_order(self, order_field):
        return self.Meta.model.objects.aggregate(
            max_order=Coalesce(Max(order_field), 0)
        )["max_order"]

    def _get_ordering_field_name(self):
        try:
            field_name = self.Meta.model._meta.ordering[0]
            if field_name[0] == "-":
                field_name = field_name[1:]
            return field_name
        except (AttributeError, IndexError):
            return None

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(
            field_name, model_field
        )
        if (
            isinstance(model_field, ORDERING_ACCEPTED_FIELDS)
            and field_name == self._get_ordering_field_name()
        ):
            field_kwargs["default"] = self.get_max_order(field_name) + 1
        return field_class, field_kwargs


class JSONFieldPatchMixin:
    """
    This mixin allows to patch JSONField values during partial updates.
    This means that, if a JSONField is present in the request and the requsest uses PATCH method,
    the serializer will merge the new data with the old one.
    """

    def is_json_field(self, attr, value, info):
        return (
            attr in info.fields
            and isinstance(info.fields[attr], DjangoJSONField)
            and isinstance(value, dict)
        )

    def update(self, instance, validated_data):
        if self.partial:
            info = model_meta.get_field_info(instance)
            for attr, value in validated_data.items():
                if self.is_json_field(attr, value, info):
                    validated_data[attr] = dict_merge(
                        getattr(instance, attr, {}), value
                    )
        return super().update(instance, validated_data)


class NestMixin:
    """
    This mixin automatically creates nested serializers for relational fields.
    The depth of the nesting can be set using the "depth" attribute of the Meta class.
    If the depth is not set, the serializer will use the value coming from the settings.

    CAMOMILLA = { "API": {"NESTING_DEPTH": 10} }
    """

    def __init__(self, *args, **kwargs):
        self._depth = kwargs.pop("depth", None)
        return super().__init__(*args, **kwargs)

    def build_nested_field(self, field_name, relation_info, nested_depth):
        return self.build_relational_field(field_name, relation_info, nested_depth)

    def build_relational_field(
        self, field_name, relation_info, nested_depth=settings.API_NESTING_DEPTH + 1
    ):
        nested_depth = nested_depth if self._depth is None else self._depth
        field_class, field_kwargs = super().build_relational_field(
            field_name, relation_info
        )
        if (
            field_class is RelatedField and nested_depth > 1
        ):  # stop recursion one step before the jump :P
            field_kwargs["serializer"] = build_standard_model_serializer(
                relation_info[1], nested_depth - 1
            )
        return field_class, field_kwargs


class AbstractPageMixin(serializers.ModelSerializer):
    """
    This mixin is needed to serialize AbstractPage models.
    It provides permalink validation and some extra fields serialization.

    Use it as a base class for your serializer if you need to serialize custom AbstractPage models.
    """

    breadcrumbs = serializers.SerializerMethodField()
    routerlink = serializers.CharField(read_only=True)
    template = serializers.SerializerMethodField()

    def get_template(self, instance: "AbstractPage"):
        return instance.get_template_path()

    def get_breadcrumbs(self, instance: "AbstractPage"):
        return instance.breadcrumbs

    LANG_PERMALINK_FIELDS = [
        build_localized_fieldname("permalink", lang)
        for lang in AVAILABLE_LANGUAGES
        if settings.ENABLE_TRANSLATIONS
    ]

    @property
    def translation_fields(self):
        return super().translation_fields + ["permalink"]

    def get_default_field_names(self, *args):
        from camomilla.contrib.rest_framework.serializer import RemoveTranslationsMixin

        if RemoveTranslationsMixin in self.__class__.__bases__:  # noqa: E501
            return super().get_default_field_names(*args)
        return (
            [f for f in super().get_default_field_names(*args) if f != "url_node"]
            + self.LANG_PERMALINK_FIELDS
            + ["permalink"]
        )

    def build_field(self, field_name, info, model_class, nested_depth):
        if field_name in self.LANG_PERMALINK_FIELDS + ["permalink"]:
            return serializers.CharField, {
                "source": "url_node.%s" % field_name,
                "read_only": True,
            }
        return super().build_field(field_name, info, model_class, nested_depth)

    def get_validators(self):
        return super().get_validators() + [UniquePermalinkValidator()]



from rest_framework import routers

from django.urls import path, include
from camomilla.views.base import BaseModelViewset
from camomilla.serializers.base import BaseModelSerializer

router = routers.DefaultRouter()
urlpatterns = []


def register(
    base_serializer=BaseModelSerializer,
    base_viewset=BaseModelViewset,
    serializer_meta={},
    viewset_attrs={},
    filters=None,
):
    """
    Register a model to the API.
    :param base_serializer: The base serializer to use for the model.
    :param base_viewset: The base viewset to use for the model.
    :param serializer_meta: The meta class to use for the serializer.
    :param viewset_attrs: The attributes to add to the viewset.
    :param filters: The filters to apply to the queryset.
    :return: The model.
    """

    def inner(model):
        base_meta = {
            "model": model,
            "fields": "__all__",
        }
        if "exclude" in serializer_meta:
            base_meta.pop("fields")
        serializer = type(
            f"{model.__name__}Serializer",
            (base_serializer,),
            {
                "Meta": type(
                    "Meta",
                    (),
                    {
                        **base_meta,
                        **serializer_meta,
                    },
                )
            },
        )

        viewset = type(
            f"{model.__name__}ViewSet",
            (base_viewset,),
            {
                "get_queryset": lambda self: model.objects.all()
                if filters is None
                else model.objects.filter(**filters),
                "serializer_class": serializer,
                **viewset_attrs,
            },
        )

        model_path = "".join(
            [
                "-" + character.lower()
                if character.isupper() and index > 0
                else character
                for index, character in enumerate(model.__name__)
            ]
        ).lstrip("-")

        router.register(
            f"{model_path.replace(' ', '_').lower().lower()}",
            viewset,
            f"{model.__name__.lower()}_api",
        )
        urlpatterns.append(path("", include(router.urls)))
        return model

    return inner



from modeltranslation.translator import TranslationOptions, register

from camomilla.models import Article, Content, Media, Page, Tag, UrlNode, Menu


class SeoMixinTranslationOptions(TranslationOptions):
    fields = (
        "title",
        "description",
        "og_description",
        "og_title",
        "og_type",
        "og_url",
        "canonical",
    )


class AbstractPageTranslationOptions(SeoMixinTranslationOptions):
    fields = ("breadcrumbs_title", "slug", "status", "indexable", "template_data")


@register(Article)
class ArticleTranslationOptions(AbstractPageTranslationOptions):
    fields = ("content",)


@register(Tag)
class TagTranslationOptions(TranslationOptions):
    fields = ("name",)


@register(Content)
class ContentTranslationOptions(TranslationOptions):
    fields = ("content",)
    empty_values = {"title": None}


@register(Media)
class MediaTranslationOptions(TranslationOptions):
    fields = ("title", "alt_text", "description")


@register(Page)
class PageTranslationOptions(AbstractPageTranslationOptions):
    pass


@register(UrlNode)
class UrlNodeTranslationOptions(TranslationOptions):
    fields = ("permalink",)


@register(Menu)
class MenuTranslationOptions(TranslationOptions):
    fields = ("nodes",)



from inspect import isclass
from typing import Any, Dict, Tuple, get_origin

import pydantic._internal._model_construction
from django.db.models import Model as DjangoModel
from pydantic import BaseModel as PyDBaseModel
from pydantic import Field, model_validator
from typing_extensions import Annotated

from .fields import ForeignKey, QuerySet
from .utils import get_type, map_method_aliases


class BaseModelMeta(pydantic._internal._model_construction.ModelMetaclass):
    def __new__(
        mcs, name: str, bases: Tuple[type], namespaces: Dict[str, Any], **kwargs
    ):
        annotations: dict = namespaces.get("__annotations__", {})
        for base in bases:
            for base_ in base.__mro__:
                if base_ is PyDBaseModel:
                    break
                annotations.update(base_.__annotations__)

        for field in annotations:
            annotation = annotations[field]
            origin = get_origin(annotation)
            if isclass(annotation) and issubclass(annotation, DjangoModel):
                annotations[field] = ForeignKey[annotation]
            elif isclass(origin) and issubclass(origin, QuerySet):
                annotations[field] = Annotated[
                    annotation,
                    Field(default_factory=get_type(annotation)._default_manager.none),
                ]
        namespaces["__annotations__"] = annotations
        return map_method_aliases(
            super().__new__(mcs, name, bases, namespaces, **kwargs)
        )


class BaseModel(PyDBaseModel, metaclass=BaseModelMeta):
    @model_validator(mode="before")
    @classmethod
    def build_cache(cls, data: Any) -> Any:
        from camomilla.structured.cache import CacheBuilder

        return CacheBuilder.from_model(cls).inject_cache(data)



import urllib.parse

from django.apps import apps
from django.conf import settings
from django.http import Http404
from django.urls import resolve, reverse
from django.utils.translation import activate, get_language

from camomilla.exceptions import NeedARedirect


def is_page(model):
    from camomilla.models import AbstractPage

    return next(
        (True for base in model.__bases__ if issubclass(base, AbstractPage)), False
    )


def get_host_url(request):
    if request:
        return "{0}://{1}".format(request.scheme, request.META["HTTP_HOST"])


def get_complete_url(request, url, language=""):
    prefix = getattr(settings, "PREFIX_DEFAULT_LANGUAGE", False)
    if language == settings.LANGUAGE_CODE and not prefix:
        language = ""
    i18n_url = urllib.parse.urljoin(language + "/", url)
    complete_url = urllib.parse.urljoin(get_host_url(request), i18n_url)
    return complete_url


def get_page(
    request,
    identifier="404",
    lang="",
    model_page=None,
    attr_page="identifier",
    model_content=None,
):
    if not model_content:
        model_content = apps.get_model(app_label="camomilla", model_name="Content")
    if not model_page:
        model_page = apps.get_model(app_label="camomilla", model_name="Page")
    if not lang:
        lang = get_language()
    kwargs = {attr_page: identifier}
    try:
        page = model_page.objects.prefetch_related("contents").get(**kwargs)
    except model_page.DoesNotExist:
        page, _ = model_page.objects.get_or_create(**kwargs)
    page = compile_seo(request, page, lang)
    return page


def get_seo_model(request, model, **params):
    seo_obj = find_or_redirect(request, model, **params)
    return compile_seo(request, seo_obj, get_language())


def compile_seo(request, seo_obj, lang=""):
    if not seo_obj.og_title:
        seo_obj.og_title = seo_obj.title
    if not seo_obj.og_description:
        seo_obj.og_description = seo_obj.description
    permalink = request.path
    if not seo_obj.permalink:
        seo_obj.permalink = permalink
    if not seo_obj.canonical:
        seo_obj.canonical = get_complete_url(request, permalink, lang)
    else:
        seo_obj.canonical = get_complete_url(request, seo_obj.canonical, lang)
    if not seo_obj.og_url:
        seo_obj.og_url = get_complete_url(request, permalink, lang)
    else:
        seo_obj.og_url = get_complete_url(request, seo_obj.og_url, lang)
    return seo_obj


def find_or_redirect(request, obj_class, **kwargs):
    try:
        obj = obj_class.objects.get(**kwargs)
        return obj
    except obj_class.DoesNotExist:
        cur_language = get_language()
        for language in settings.LANGUAGES:
            try:
                activate(language[0])
                obj = obj_class.objects.get(**kwargs)
                activate(cur_language)
                obj = obj_class.objects.get(pk=obj.pk)
                args = []
                for kwarg_key, kwarg_val in kwargs.items():
                    args.append(getattr(obj, kwarg_key))
                url_name = resolve(request.path_info).url_name
                language_path = reverse(url_name, args=args)
                raise NeedARedirect(language_path)
            except obj_class.DoesNotExist:
                pass
        activate(cur_language)
        raise Http404()



from pydantic import TypeAdapter, ValidationError
from rest_framework import serializers
from rest_framework.utils import model_meta
from typing import TYPE_CHECKING, Any, Union

from camomilla.structured.utils import pointed_setter

if TYPE_CHECKING:
    from camomilla.structured import BaseModel


class StructuredJSONField(serializers.JSONField):
    """
    This field allows to serialize and deserialize structured data.
    """

    schema: Union["BaseModel", TypeAdapter] = None

    def __init__(self, **kwargs):
        self.schema = kwargs.pop("schema", None)
        super().__init__(**kwargs)

    def bind(self, field_name, parent):
        if self.schema is None and isinstance(parent, serializers.ModelSerializer):
            info = model_meta.get_field_info(parent.Meta.model)
            field = info.fields[field_name]
            self.schema = field.schema
            self.many = field.many
            self.json_schema = field.schema.json_schema()
        super().bind(field_name, parent)

    def to_representation(self, instance: Union["BaseModel", list["BaseModel"]]):
        if isinstance(instance, list) and self.many:
            return super().to_representation(
                self.schema.dump_python(instance, exclude_unset=True)
            )
        return super().to_representation(instance.model_dump(exclude_unset=True))

    def to_internal_value(self, data: Union[list, dict]):
        try:
            return self.schema.validate_python(super().to_internal_value(data))
        except ValidationError as e:
            drf_error: Union[list, dict[str, Any]] = [] if self.many else {}
            for error in e.errors():
                pointed_setter(
                    drf_error, ".".join([str(x) for x in error["loc"]]), [error["msg"]]
                )
            raise serializers.ValidationError(drf_error)



from camomilla.models.page import UrlNode
from camomilla.serializers.mixins import AbstractPageMixin
from camomilla.models import Content, Page
from camomilla.serializers.base import BaseModelSerializer
from rest_framework import serializers

from camomilla.serializers.utils import (
    build_standard_model_serializer,
    get_standard_bases,
)


class ContentSerializer(BaseModelSerializer):
    class Meta:
        model = Content
        fields = "__all__"


class PageSerializer(AbstractPageMixin, BaseModelSerializer):
    class Meta:
        model = Page
        fields = "__all__"


class BasicUrlNodeSerializer(BaseModelSerializer):
    is_public = serializers.SerializerMethodField()
    status = serializers.SerializerMethodField()
    indexable = serializers.SerializerMethodField()

    class Meta:
        model = UrlNode
        fields = ("id", "permalink", "status", "indexable", "is_public")

    def get_is_public(self, instance: UrlNode):
        return instance.page.is_public

    def get_status(self, instance: UrlNode):
        return instance.page.status

    def get_indexable(self, instance: UrlNode):
        return instance.page.indexable


class UrlNodeSerializer(BasicUrlNodeSerializer):
    alternates = serializers.SerializerMethodField()

    def get_alternates(self, instance: UrlNode):
        return instance.page.alternate_urls()

    def to_representation(self, instance: UrlNode):
        model_serializer = build_standard_model_serializer(
            instance.page.__class__,
            depth=10,
            bases=(AbstractPageMixin,) + get_standard_bases(),
        )
        return {
            **super().to_representation(instance),
            **model_serializer(instance.page, context=self.context).data,
        }

    class Meta:
        model = UrlNode
        fields = "__all__"



from rest_framework import serializers

from ...contrib.rest_framework.serializer import TranslationsMixin
from ..fields import FieldsOverrideMixin
from ..mixins import (
    JSONFieldPatchMixin,
    NestMixin,
    OrderingMixin,
    SetupEagerLoadingMixin,
)


class BaseModelSerializer(
    NestMixin,
    FieldsOverrideMixin,
    JSONFieldPatchMixin,
    OrderingMixin,
    SetupEagerLoadingMixin,
    TranslationsMixin,
    serializers.ModelSerializer,
):
    """
    This is the base serializer for all the models.
    It adds support for:
    - nesting translations fields under a "translations" field
    - overriding related fields with auto-generated serializers
    - patching JSONField
    - ordering
    - eager loading
    """

    pass


__all__ = [
    "BaseModelSerializer",
]



from django.conf import settings
from django.db import migrations, connection


class KeepTranslationsMixin:
    """
    This mixin make it possible to keep translations when migrating from django-hvad to modeltranslation and viceversa.
    To use it, you have to add a dictionary to your migration class called "keep_translations".
    The dictionary must have model paths as keys and a list of fields to keep as values.


    Example:
    ```python
    class Migration(KeepTranslationsMixin, migrations.Migration):
        keep_translations = {
            "app.Model": ("field1", "field2", "field3")
        }
    ```
    """

    _saved_data_from_plain = {}
    language_codes = dict(getattr(settings, "LANGUAGES", {})).keys()

    def is_operation_legit(self, o):
        return not (
            isinstance(o, migrations.RemoveField)
            and o.name == "master"
            and o.model_name.endswith("translation")
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.operations = [o for o in self.operations if self.is_operation_legit(o)]
        self.operations.insert(
            0, migrations.RunPython(self._getDataFromHvad, self._restoreDataToHvad)
        )
        self.operations.append(
            migrations.RunPython(
                self._restoreDataToModelTranslation, self._getDataFromModelTranslation
            )
        )

    def _getDataFromHvad(self, apps, schemaeditor):
        for modelPath, fields in self.keep_translations.items():
            Model = apps.get_model(*modelPath.split("."))
            table = Model._meta.db_table + "_translation"
            if "language_code" not in fields:
                fields = ("language_code",) + fields
            with connection.cursor() as cursor:
                cursor.execute("SELECT master_id FROM {0};".format(table))
                masters = list(set(cursor.fetchall()))
                for master in masters:
                    cursor.execute(
                        "SELECT {0} FROM {1} WHERE master_id={2};".format(
                            ",".join(fields), table, master[0]
                        )
                    )
                    rows = cursor.fetchall()
                    self._saved_data_from_plain[
                        modelPath
                    ] = self._saved_data_from_plain.get(modelPath, {})
                    self._saved_data_from_plain[modelPath][
                        master[0]
                    ] = self._saved_data_from_plain[modelPath].get(master[0], [])
                    for row in rows:
                        self._saved_data_from_plain[modelPath][master[0]].append(
                            dict(zip(fields, row))
                        )

    def _getDataFromModelTranslation(self, apps, schemaeditor):
        for modelPath, fields in self.keep_translations.items():
            Model = apps.get_model(*modelPath.split("."))
            table = Model._meta.db_table
            for lang in self.language_codes:
                t_fields = ("id",) + tuple(
                    "{0}_{1}".format(f, lang) for f in fields if f != "id"
                )
                with connection.cursor() as cursor:
                    cursor.execute(
                        "SELECT {0} FROM {1}".format(
                            ",".join(t_fields),
                            table,
                        )
                    )
                    rows = cursor.fetchall()
                    self._saved_data_from_plain[
                        modelPath
                    ] = self._saved_data_from_plain.get(modelPath, [])
                    for row in rows:
                        row_data = dict(zip(("master_id", *fields), row))
                        row_data.update({"language_code": lang})
                        self._saved_data_from_plain[modelPath].append(row_data)

    def _restoreDataToModelTranslation(self, apps, schemaeditor):
        for key, master_dict in self._saved_data_from_plain.items():
            Model = apps.get_model(*key.split("."))
            for pk, translations in master_dict.items():
                try:
                    obj = Model.objects.get(pk=pk)
                except Model.DoesNotExist:
                    continue
                for translation in translations:
                    lang = translation.pop("language_code")
                    for attr, value in translation.items():
                        setattr(obj, "{0}_{1}".format(attr, lang), value)
                obj.save()

    def _restoreDataToHvad(self, apps, schemaeditor):
        for key, rows in self._saved_data_from_plain.items():
            Model = apps.get_model(*key.split("."))
            table = Model._meta.db_table + "_translation"
            for row in rows:
                with connection.cursor() as cursor:
                    print(row)
                    cursor.execute(
                        "INSERT INTO {0} ({1}) VALUES ({2});".format(
                            table,
                            ",".join(row.keys()),
                            ",".join(
                                [
                                    "'{}'".format(v).replace("'None'", "NULL")
                                    for v in row.values()
                                ]
                            ),
                        )
                    )



def dict_merge(dct: dict, merge_dct: dict) -> dict:
    for k, v in merge_dct.items():
        if k in dct and isinstance(dct[k], dict) and isinstance(v, dict):  # noqa
            dict_merge(dct[k], v)
        else:
            dct[k] = v
    return dct



from django.contrib.auth import get_user_model, login, logout
from django.contrib.auth.models import Permission
from django.db.models import Q
from rest_framework.authtoken.models import Token
from rest_framework.authtoken.views import ObtainAuthToken
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from camomilla.permissions import CamomillaBasePermissions, CamomillaSuperUser, ReadOnly
from camomilla.serializers import (
    PermissionSerializer,
    UserProfileSerializer,
    UserSerializer,
)
from camomilla.views.base import BaseModelViewset


class CamomillaObtainAuthToken(ObtainAuthToken):
    def post(self, request, *args, **kwargs):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data["user"]
        token, _ = Token.objects.get_or_create(user=user)
        return Response({"token": token.key})


class UserViewSet(BaseModelViewset):
    queryset = get_user_model().objects.all()
    serializer_class = UserSerializer
    model = get_user_model()
    permission_classes = (CamomillaSuperUser | CamomillaBasePermissions,)

    @action(detail=False, methods=["get", "put"], permission_classes=(IsAuthenticated,))
    def current(self, request):
        user = request.user
        if request.method == "PUT":
            serialized_user = UserProfileSerializer(
                user, data=request.data, partial=True
            )
            if serialized_user.is_valid(raise_exception=True):
                user = serialized_user.save()
        return Response(UserProfileSerializer(user, context={"request": request}).data)

    @action(detail=True, methods=["post"])
    def kickout(self, request, pk=None):
        user = self.get_object()
        try:
            user.auth_token.delete()
        except Exception:
            pass
        return Response({})


class PermissionViewSet(BaseModelViewset):
    queryset = Permission.objects.all()
    serializer_class = PermissionSerializer
    model = Permission
    permission_classes = (CamomillaSuperUser | ReadOnly,)
    http_method_names = ["get", "put", "options", "head"]

    def get_queryset(self):
        permissions = Permission.objects.filter(
            Q(content_type__app_label__contains="camomilla")
            | Q(content_type__app_label__contains="plugin_")
            | Q(content_type__model="token")
            | Q(content_type__model="user")
        )
        return permissions


class CamomillaAuthLogin(ObtainAuthToken):
    def post(self, request, *args, **kwargs):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data["user"]
        login(request, user)
        return Response({"detail": "Logged in"})


class CamomillaAuthLogout(APIView):
    def get(self, request, *args, **kwargs):
        if request.user:
            logout(request)
        return Response({"detail": "Logged out"})



from functools import cached_property
from typing import Iterable, Union
from django.http import QueryDict

from modeltranslation import settings as mt_settings
from modeltranslation.translator import NotRegistered, translator
from modeltranslation.utils import build_localized_fieldname
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from camomilla.utils.getters import pointed_getter
from camomilla.utils.translation import is_translatable


TRANS_ACCESSOR = "translations"


def plain_to_nest(data, fields, accessor=TRANS_ACCESSOR):
    """
    This function transforms a plain dictionary with translations fields (es. {"title_en": "Hello"})
    into a dictionary with nested translations fields (es. {"translations": {"en": {"title": "Hello"}}}).
    """
    trans_data = {}
    for lang in mt_settings.AVAILABLE_LANGUAGES:
        lang_data = {}
        for field in fields:
            trans_field_name = build_localized_fieldname(field, lang)
            if trans_field_name in data:
                lang_data[field] = data.pop(trans_field_name)
        if lang_data.keys():
            trans_data[lang] = lang_data
    if trans_data.keys():
        data[accessor] = trans_data
    return data


def nest_to_plain(data: Union[dict, QueryDict], fields: list[str], accessor=TRANS_ACCESSOR):
    """
    This function is the inverse of plain_to_nest.
    It transforms a dictionary with nested translations fields (es. {"translations": {"en": {"title": "Hello"}}})
    into a plain dictionary with translations fields (es. {"title_en": "Hello"}).
    """
    if isinstance(data, QueryDict):
        data = data.dict()
    translations = data.pop(accessor, {})
    for lang in mt_settings.AVAILABLE_LANGUAGES:
        nest_trans = translations.pop(lang, {})
        for k in fields:
            data.pop(k, None)  # this removes all trans field without lang
            if k in nest_trans:
                # this saves on the default field the default language value
                if lang == mt_settings.DEFAULT_LANGUAGE:
                    data[k] = nest_trans[k]
                key = build_localized_fieldname(k, lang)
                data[key] = data.get(key, nest_trans[k])
    return data


class TranslationsMixin(serializers.ModelSerializer):
    """
    This mixin adds support for modeltranslation fields.
    It automatically nests all translations fields (es. title_en) under a "translations" field.

    This means that, in representation, the serializer will transform:
    `{"title_en": "Hello", "title_it": "Ciao"}` -> `{"translations": {"en": {"title": "Hello"}, "it": {"title": "Ciao"}}`

    While in deserialization, the serializer will transform:
    `{"translations": {"en": {"title": "Hello"}, "it": {"title": "Ciao"}}` -> `{"title_en": "Hello", "title_it": "Ciao"}`
    """

    @cached_property
    def translation_fields(self) -> list[str]:
        try:
            return translator.get_options_for_model(self.Meta.model).get_field_names()
        except NotRegistered:
            return []

    @property
    def _writable_fields(self) -> Iterable[serializers.Field]:
        for field in super()._writable_fields:
            if field.field_name not in self.translation_fields:
                yield field

    def to_internal_value(self, data):
        if self.translation_fields:
            nest_to_plain(data, self.translation_fields)
        return super().to_internal_value(data)

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        if self.translation_fields:
            plain_to_nest(representation, self.translation_fields)
        return representation

    def run_validation(self, *args, **kwargs):
        try:
            return super().run_validation(*args, **kwargs)
        except ValidationError as ex:
            if self.translation_fields:
                plain_to_nest(ex.detail, self.translation_fields)
            raise ValidationError(detail=ex.detail)

    @property
    def is_translatable(self):
        return is_translatable(pointed_getter(self, "Meta.model"))


class RemoveTranslationsMixin(serializers.ModelSerializer):
    """
    This mixin removes all translations fields (es. title_en) from the serializer.
    It's useful when you want to create a serializer that doesn't need to include all translations fields.

    If request is passed in context, this serializer becomes aware of the query parameter "included_translations".
    If the value is "all", all translations fields are included.
    If the value is a comma separated list of languages (es. "en,it"), only the specified translations fields are included.
    """

    @cached_property
    def translation_fields(self):
        try:
            return translator.get_options_for_model(self.Meta.model).get_field_names()
        except NotRegistered:
            return []

    def get_default_field_names(self, declared_fields, model_info):
        request = self.context.get("request", False)
        included_translations = request and request.GET.get(
            "included_translations", False
        )
        if included_translations == "all":
            return super().get_default_field_names(declared_fields, model_info)
        elif included_translations is not False:
            included_translations = included_translations.split(",")
        else:
            included_translations = []

        field_names = super().get_default_field_names(declared_fields, model_info)
        for lang in mt_settings.AVAILABLE_LANGUAGES:
            if lang not in included_translations:
                for field in self.translation_fields:
                    localized_fieldname = build_localized_fieldname(field, lang)
                    if localized_fieldname in field_names:
                        field_names.remove(localized_fieldname)
        return field_names



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.http import Http404
from django.shortcuts import get_object_or_404
from rest_framework.decorators import action
from rest_framework.response import Response

from camomilla.models import AbstractPage, Menu
from camomilla.models.page import UrlNode
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import ContentTypeSerializer, MenuSerializer
from camomilla.serializers.page import BasicUrlNodeSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.decorators import active_lang

from django.utils.translation import get_language


class MenuViewSet(BaseModelViewset):
    queryset = Menu.objects.all()
    serializer_class = MenuSerializer
    permission_classes = (CamomillaBasePermissions,)
    model = Menu

    lookup_field = "key"

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        assert lookup_url_kwarg in self.kwargs, (
            "Expected view %s to be called with a URL keyword argument "
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            "attribute on the view correctly."
            % (self.__class__.__name__, lookup_url_kwarg)
        )
        filter = Q(**{self.lookup_field: self.kwargs[lookup_url_kwarg]})
        if (
            isinstance(self.kwargs[lookup_url_kwarg], str)
            and self.kwargs[lookup_url_kwarg].isnumeric()
            or isinstance(self.kwargs[lookup_url_kwarg], int)
        ):
            filter |= Q(pk=self.kwargs[lookup_url_kwarg])

        obj = get_object_or_404(queryset, filter)
        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    @action(detail=False, methods=["get"], url_path="page_types")
    def page_types(self, request, *args, **kwargs):
        return Response(
            ContentTypeSerializer(
                [
                    model
                    for model in ContentType.objects.all()
                    if model.model_class()
                    and issubclass(model.model_class(), AbstractPage)
                ],
                many=True,
            ).data
        )

    @action(detail=False, methods=["get"], url_path=r"page_types/(?P<content_id>\w+)")
    def page_type_instances(self, request, content_id, *args, **kwargs):
        content_type = ContentType.objects.filter(pk=content_id).first()
        if not content_type or not issubclass(content_type.model_class(), AbstractPage):
            raise Http404("No object matches the given query.")
        return Response(
            [
                {"id": obj.pk, "name": str(obj), "url_node_id": obj.url_node.pk}
                for obj in content_type.model_class().objects.exclude(
                    url_node__isnull=True
                )
            ]
        )

    @active_lang()
    @action(detail=False, methods=["get"], url_path="search_urlnode")
    def search_urlnode(self, request, *args, **kwargs):
        url_node = request.GET.get("q", "")
        qs = UrlNode.objects.filter(permalink__icontains=url_node).order_by("permalink")
        print(get_language())
        return Response(BasicUrlNodeSerializer(qs, many=True).data)



from django import template
from django.utils.translation import get_language


register = template.Library()


@register.filter(name="filter_content")
def filter_content(page, args):
    try:
        content = page.contents.get(identifier=args)
    except page.contents.model.DoesNotExist:
        content, _ = page.contents.get_or_create(identifier=args)
    return content


@register.filter(name="alternate_urls")
def alternate_urls(page, request):
    alternates = page.alternate_urls(request)
    return alternates.get("alternate_urls", alternates).items()


@register.filter(name="strip_lang")
def strip_lang(value, lang=get_language()):
    return "/%s" % value.lstrip("/%s/" % lang)



import json

from django.http import JsonResponse
from rest_framework.decorators import action

from camomilla.models import Content
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import ContentSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.mixins import BulkDeleteMixin, GetUserLanguageMixin


class ContentViewSet(GetUserLanguageMixin, BulkDeleteMixin, BaseModelViewset):
    queryset = Content.objects.all()
    serializer_class = ContentSerializer
    model = Content
    permission_classes = (CamomillaBasePermissions,)

    @action(detail=True, methods=["get", "patch"])
    def djsuperadmin(self, request, pk):
        try:
            content = Content.objects.get(pk=pk)
        except Content.DoesNotExist:
            content, _ = Content.objects.get_or_create(pk=pk)
        if request.method == "GET":
            return JsonResponse({"content": content.content})
        if request.method == "PATCH":
            data = json.loads(request.body)
            content_data = data["content"]
            content.content = content_data
            content.save()
            return JsonResponse({"content": content_data})



from django.db import models


class SeoMixin(models.Model):
    title = models.CharField(max_length=200, null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    og_description = models.TextField(blank=True, null=True)
    og_title = models.CharField(max_length=200, blank=True, null=True)
    og_type = models.CharField(max_length=200, blank=True, null=True)
    og_url = models.CharField(max_length=200, blank=True, null=True)
    canonical = models.CharField(max_length=200, blank=True, null=True)
    og_image = models.ForeignKey(
        "camomilla.Media",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="%(app_label)s_%(class)s_og_images",
    )

    class Meta:
        abstract = True


class MetaMixin(models.Model):
    meta = models.JSONField(default=dict, null=False, blank=True)

    def get_meta(self, key, default=None):
        return self.meta.get(key, default)

    def update_meta(self, key, value):
        self.meta[key] = value
        super(MetaMixin, self).save(update_fields=["meta"])

    def delete_meta(self, key):
        del self.meta[key]
        super(MetaMixin, self).save(update_fields=["meta"])

    class Meta:
        abstract = True
